#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <values.h>
#include <time.h>
#include <sys/time.h>
#include <cpgplot.h>
//#include <complex.h>
#include <fftw3.h>

//#define MALLOC fftwf_malloc
#define MALLOC malloc

#define MIN(x,y)   ((x) < (y) ? (x) : (y))

#define MAX(x,y)   ((x) > (y) ? (x) : (y))


double read_head(int file, struct tm *tm);
int autocorr(char *data, long *lags, int nlags, int nread);

#define BLOCKSIZE 1024*1024

#define RATE 32
#define MAXSTR 256

typedef enum {VLBA, AT} encoding_type;

double tim(void) {
  struct timeval tv;
  double t;

  gettimeofday(&tv, NULL);
  t = (double)tv.tv_sec + (double)tv.tv_usec/1000000.0;

  return t;
}

#define MAXCHAN 8

int main (int argc, char * const argv[]) {
  int nfile, i, j, k, l, file, status, nread, opt, first, memsize, tmp;
  int naver, readshort, nfft, nx, ny, swapped, maxchan, nplot;
  long *lags[MAXCHAN];
  register int samp;
  register char *mptr;
  char msg[MAXSTR+1];
  char *data[2][MAXCHAN], lookup[4];
  unsigned char  *mem;
  unsigned short *mem16;
  off_t offset;
  float *in[MAXCHAN], *plotspec, *xvals, min, max, delta, sampperbyte;
  fftwf_complex *out[MAXCHAN];
  double time, *spectrum[MAXCHAN], t0, t1, tt;
  long long lcount[4], tcount;
  struct tm tm;
  fftwf_plan plans[MAXCHAN]; /* Do up to MAXCHAN subbands at the same time */

  int verbose;       /* Prints whats happening */
  int bits;          /* Total number of bits/sample */
  int MHz64;         /* 64 MHz mode (4 consecutive samples/byte) */
  int nskip;         /* Number of bytes to skip at start of file */
  int nlags;         /* Number of lags to correlate */
  int chans[MAXCHAN];/* Channels tocorrelate */
  int bchan;         /* First spectral point to plot */
  int echan;         /* Last spectral point to plot */
  encoding_type encoding; /* VLBA or AT bit encoding (offset binary or sign/mag) */
  char pgdev[MAXSTR+1] = "/xs"; /* Pgplot device to use */

  int chan1, chan2;

  struct option options[] = {
    {"nlags", 1, 0, 'n'},
    {"skip", 1, 0, 's'},
    {"bits", 1, 0, 'b'},
    {"chan", 1, 0, 'c'},
    {"if1", 1, 0, '1'},
    {"if2", 1, 0, '2'},
    {"1", 1, 0, '1'},
    {"2", 1, 0, '2'},
    {"sp1", 1, 0, 'x'},
    {"sp2", 1, 0, 'y'},
    {"device", 1, 0, 'd'},
    {"pgdev", 1, 0, 'd'},
    {"vlba", 0, 0, 'v'},
    {"wide", 0, 0, 'w'},
    {"64", 0, 0, 'w'},
    {"verbose", 0, 0, 'V'},
    {"V", 0, 0, 'V'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  verbose = 0;
  bits = 8;
  nskip = -1;
  nlags = 128;
  encoding = AT;
  MHz64 = 0;
  chan1 = -1;
  chan2 = -1;
  for (k=0;k<MAXCHAN;k++) chans[k]=0;
  bchan = -1;
  echan = -1;

  nfft = 0;
  /* Read command line options */
  while (1) {
    opt = getopt_long_only(argc, argv, "", options, NULL);
    if (opt==EOF) break;

    switch (opt) {
      
    case 'b':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -bits option %s\n", optarg);
      else 
	bits = tmp;
      break;

    case 'c':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -chan option %s\n", optarg);
      else {
	if (tmp>MAXCHAN) {
	  fprintf(stderr, "Chan %s too large\n", optarg);
	} else if (tmp<=0) {
	  fprintf(stderr, "Chan %s too small\n", optarg);
	} else {
	  chans[tmp-1] = 1;
	}
      }
      break;

    case '1':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -1chan option %s\n", optarg);
      else 
	chan1 = tmp-1;
      break;

    case '2':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -2chan option %s\n", optarg);
      else 
	chan2 = tmp;
      break;

    case 'x':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -x option %s\n", optarg);
      else 
	bchan = tmp;
      break;

    case 'y':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -z option %s\n", optarg);
      else 
	echan = tmp;
      break;

    case 's':
      fprintf(stderr, "Skip option not yet supported\n");
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -skip option %s\n", optarg);
      else 
	nskip = tmp;
      break;

    case 'n':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -nlags option %s\n", optarg);
      else 
	nlags = tmp;
      break;

    case 'd':
      if (strlen(optarg)>MAXSTR) {
	fprintf(stderr, "PGDEV option %s too long\n", optarg);
	return 0;
      }
      strcpy(pgdev, optarg);
      break;

    case 'V':
      verbose = 1;
      break;

    case 'v':
      encoding = VLBA;
      break;

    case 'w':
      MHz64 = 1;
      break;

    case 'h':
      printf("Usage: vsib_checker [options]\n");
      printf("  -b/-bits        Total number bits/sample (4/8/16)\n");
      printf("  -s/-skip <n>    Skip this many bytes at start of file\n");
      printf("  -v/-vlba        Calculate statistics for VLBA data\n");
      printf("  -w/-wide/-64    Assume 64 MHz bandwidth data\n");
      printf("  -c/-chan <n>    Channel to correlate (can specify multipl times)\n");
      printf("  -if1/-1 <n>     1st channel to correlate in range\n");
      printf("  -if2/-2 <n>     Last channel to correlate in range\n");
      printf("  -sp1 <n>        First spectral point to plot\n");
      printf("  -sp2 <n>        Last spectral point to plot\n");
      printf("  -n/-nlags <n>   # lags to correlate\n");
      printf("  -device <pgdev> Pgplot device to plot to\n");
      printf("  -V/-verbose     Verbose\n");
      printf("  -h/-help        This list\n");
      return(1);
      break;

    case '?':
    default:
      break;
    }
  }

  /* Check too high channel hasn't been selected */
  maxchan = bits/2; 
  if (MHz64) maxchan /= 4;
  for (k=maxchan; k<MAXCHAN; k++) {
    if (chans[k]) {
      fprintf(stderr, "Channel %d is too large for this mode. Ignoring\n", k+1);
      chans[k] = 0;
    }
  }

  if (chan1>0 || chan2>0) {
    if (chan1<=0) 
      chan1=1;
    else if (chan1>maxchan) {
      fprintf(stderr, "Channel %d too large for current mode. Setting to %d\n",
	      chan1, maxchan);
      chan1=maxchan;
    }

    if (chan2<=0) 
      chan2=maxchan;
    else if (chan2>maxchan) {
      fprintf(stderr, "Channel %d too large for current mode. Setting to %d\n",
	      chan2, maxchan);
      chan2=maxchan;
    }
    if (chan2<chan1) {
      tmp = chan1;
      chan1 = chan2;
      chan2 = tmp;
    }
    for (k=chan1-1;k<chan2;k++) {
      chans[k] = 1;
    }
  }

  /* Have any channels been selected? */
  nplot = 0;
  for (k=0; k<maxchan; k++) {
    if (chans[k]) {
      nplot++;
    }
  }
  if (nplot==0) {
    fprintf(stderr, "Correlating all channels\n");

    for (k=0; k<maxchan; k++) {
      chans[k] = 1;
      nplot++;
    }
  }

  if (bchan<0) bchan = 0;
  if (echan<0) echan = nlags/2-1;
  if (bchan > nlags/2-1) bchan = nlags/2-1;
  if (echan > nlags/2-1) echan = nlags/2-1;
  if (bchan==echan) {
    fprintf(stderr, "Warning: Odd channel range. Resetting\n");
    bchan = 0;
    echan = nlags/2-1;
  }
  
  swapped = 0;
  if (bchan>echan) {
    swapped = bchan;
    bchan = echan;
    echan = swapped;
    swapped = 1;
    fprintf(stderr, "Swapping frequency axis\n");
  } 
  echan++;

  sampperbyte = 8/bits; /* Bytes per integration */
  if (MHz64) sampperbyte *=4;  /* 4 samples per byte */
  printf("Samples per byte = %0.1f\n", sampperbyte);

  memsize = BLOCKSIZE;

  mem = MALLOC(memsize);
  if (mem==NULL) {
    sprintf(msg, "Trying to allocate %d byes", memsize);
    perror(msg);
    return(1);
  }

  /* Allocate some memory to unpack the data into and accumulate correlation */

  for (k=0; k<maxchan; k++) {
    if (!chans[k]) continue;
    data[0][k] = MALLOC(memsize*sampperbyte);
    if (data[0][k]==NULL) {
      perror("Allocating memory");
      exit(1);
    }

    lags[k] = MALLOC(sizeof(long)*nlags);
    if (lags[k]==NULL) {
      perror("Allocating memory");
      exit(1);
    }
    for (i=0; i<nlags; i++) {
      lags[k][i] = 0;
    }
  }

  /* Initiate fftw */

  for (k=0; k<maxchan; k++) {
    if (!chans[k]) continue;
    in[k] = MALLOC(sizeof(float) * nlags);
    if (in[k]==NULL) {
      perror("Allocating memory");
      exit(1);
    }
    out[k] = MALLOC(sizeof(fftwf_complex) * nlags/2);
    if (out[k]==NULL) {
      perror("Allocating memory");
      exit(1);
    }
    spectrum[k] = MALLOC(sizeof(double)*nlags/2);
    if (spectrum[k]==NULL) {
      perror("Allocating memory");
      exit(1);
    }
    for (j=0; j<nlags/2; j++) {
      spectrum[k][j] = 0;
    }

    plans[k] = fftwf_plan_dft_r2c_1d(nlags, in[k], out[k], 0);
  }
  naver = 0;

  /* Setup the lookup table */
  if (encoding==VLBA) {
    lookup[0] = +3;
    lookup[1] = +1;
    lookup[2] = -1;
    lookup[3] = -3;
  } else { /* AT */
    lookup[0] = +1;
    lookup[1] = -1;
    lookup[2] = +3;
    lookup[3] = -3;
  }

  tt = 0; /* Total time */

  for (nfile=optind; nfile<argc; nfile++) {

    file = open(argv[nfile], O_RDONLY|O_LARGEFILE);
    if (file==-1) {
      sprintf(msg, "Failed to open input file (%s)", argv[nfile]);
      perror(msg);
      //fftwf_free(mem);
      return(1);
    }
    printf("%s\n", argv[nfile]);

    /* Read the header */
    time = read_head(file, &tm);

    for (i=0; i<4; i++) {
      lcount[i] = 0;
    }

    offset = 0;
    first = 1;
    readshort = 0;

    while (1) {
      nread = read(file, mem, memsize);

      if (nread==0) {
	if (offset==0) {
	  fprintf(stderr, "No data read from %s\n", argv[nfile]);
	  //fftwf_free(mem);
	  close(file);
	  return(1);
	}
	break;
      } else if (nread==-1) {
	perror("Error reading file");
	close(file);
	//fftwf_free(mem);
	return(1);
      } 
      offset += nread;
      t0 = tim();

      if (nread<nlags) { /* Not read enough */
	if (readshort) {
	  /* Probably at end of file */
	  if (offset < nlags) {
	    fprintf(stderr, "Warning: Short file\n");
	    exit(1);
	  } else {
	    break; 
	  }
	} else {
	  readshort = 1; /* Try one more time, just in case */
	  continue;
	}
      }

      mptr = mem;

      /* The following loops decode the data into a set of byte arrays */

      /* Copy data into "data" array, then correlate and accumulate */
      if (MHz64) { /* 64 MHz needs to be handles seperatelt */
	if (bits==8) {
	  fprintf(stderr, "64 MHz 8 bit not supported\n");
	  exit(1);
	  
	} else if (bits==16) {
	  for (i=0; i<nread/2; i++) {
	    for (k=0; k<maxchan; k++) {
	      if (!chans[k]) continue;
	      samp = *mptr;
	      for (l=0; l<4; l++) {
		// Use pointer++ here as well?
		data[0][k][i*4+l] = lookup[(samp>>(l*2))&0x3];
		lcount[(samp>>(l*2))&0x3]++;
	      }
	      mptr++;
	    }
	  }
	}
      } else {
	if (bits==8) {
	  for (i=0; i<nread; i++) {
	    samp = *mptr;
	    for (k=0; k<maxchan; k++) {
	      if (!chans[k]) continue;
	      data[0][k][i] = lookup[(samp>>(k*2))&0x3];
	      lcount[(samp>>(k*2))&0x3]++;
	    }
	    mptr++;
	  }
	} else if (bits==16) {
	  mem16 = (unsigned short*)mem;
	  for (i=0; i<nread/2; i++) {
	    samp = *mem16;
	    for (k=0; k<maxchan; k++) {
	      if (!chans[k]) continue;
	      data[0][k][i] = lookup[(samp>>(k*2))&0x3];
	      lcount[(samp>>(k*2))&0x3]++;
	    }
	    mem16++;
	  }
	} else if (bits==4) {
	  for (i=0; i<nread; i++) {
	    samp = *mptr;
	    for (k=0; k<maxchan; k++) {
	      if (!chans[k]) continue;
	      in[k][i*2] = lookup[(samp>>(k*2))&0x3];
	      in[k][i*2+1] = lookup[(samp>>(k*2+4))&0x3];
	      lcount[(samp>>(k*2))&0x3]++;
	      lcount[(samp>>(k*2+4))&0x3]++;
	    }
	    mptr++;
	  }
	} else {
	  fprintf(stderr, "Unknown number of bits %d\n", bits);
	  exit(1);
	}
      }

      for (k=0; k<maxchan; k++) {
	if (!chans[k]) continue;
	
	autocorr(data[0][k], lags[k], nlags, nread);
	
      }
      naver++;
      t1 = tim();
      tt += t1-t0;
    }
  }

  //fftwf_execute(plans[k]);
  //nfft++;

  printf("Total computational time= %0.1f seconds for %d ffts\n", 
	 tt, nfft);

  //fftwf_free(mem);
  for (k=0; k<maxchan; k++) {
    if (!chans[k]) continue;

    for (j=0; j<nlags; j++) 
      spectrum[k][j] = data[0][k][j];
    
    //fftwf_destroy_plan(plans[k]);
    //fftwf_free(in[k]); 
    //fftwf_free(out[k]);
  }

  xvals = MALLOC(sizeof(float)*nlags);
  if (xvals==NULL) {
    sprintf(msg, "Trying to allocate %d byes", sizeof(float)*nlags);
    perror(msg);
    return(1);
  }
  for (j=0; j<nlags; j++) {
    xvals[j] = j+0.5;
  }

  /* How many plots to do */;
  ny = ceil(sqrt(nplot));
  nx = (nplot+ny-1)/ny;
  
  if (cpgopen(pgdev)<0) {
    fprintf(stderr, "Error calling PGOPEN");
    return(1);
  }
  //cpgsubp(nx, ny);
  
  plotspec = MALLOC(sizeof(float)*nlags);
  if (xvals==plotspec) {
    sprintf(msg, "Trying to allocate %d byes", sizeof(float)*nlags);
    perror(msg);
    return(1);
  }

  for (k=0; k<maxchan; k++) {
    if (!chans[k]) continue;

    max = 0;
    min = 0;

    for (j=0; j<nlags; j++) {
      plotspec[j] = lags[k][j];
      if (plotspec[j]>max) 
	  max = plotspec[j];
      else if (plotspec[j]<min) 
	  min = plotspec[j];
    }
    printf("Max = %0.f  Min = %0.f\n", max, min);

    for (j=0; j<nlags; j++) {
      plotspec[j] /= max;
	printf("%3d %.5f\n", j, plotspec[j]);
    }
    min /= max;

    delta = (max-min)*0.05;
    if (min<0) min -= delta;
    max += delta;

    //cpgsci(7);
    //if (swapped) 
      //cpgenv(echan,bchan,min,max,0,0);
	//else 
      //cpgenv(bchan,echan,min,max,0,0);
    sprintf(msg, "Channel %d", k+1);
    //cpglab("Channel", "Correlation", msg);
    //cpgsci(2);
    //cpgline(echan-bchan, xvals, &plotspec[bchan]);

    //fftwf_free(spectrum[k]);
    //fftwf_free(plotspec[k]);

  }
  //fftwf_free(xvals);

  tcount = 0;
  for (i=0; i<4; i++) {
    tcount += lcount[i];
  }
  printf("Total counts %.1fe6\n", tcount/1e6);
  for (i=0; i<4; i++) {
    printf(" %5.1f", lcount[i]/(double)tcount*100);
  }
  printf("\n");

  //cpgend();

  return(1);
}

int autocorr(char *data, long *lags, int nlags, int nread) {
  char *p1, *p2, *p3;
  long *lagptr;

  p1 = data+nlags/2-1;
  while (p1<data+nread-nlags/2) {
    p2 =  p1-nlags/2+1;
    p3 = p2+nlags;
    lagptr = lags;
    while (p2<p3) {
      *lagptr += *p1 * *p2; // The actual correlation!
      p2++;
      lagptr++;
    }
    p1++;
  }

  return 0;
}
